"use strict";

exports.__esModule = true;

var _trimEnd = require("lodash/trimEnd");

var _trimEnd2 = _interopRequireDefault(_trimEnd);

var _tokenTypes = require("./tokenTypes");

var _tokenTypes2 = _interopRequireDefault(_tokenTypes);

var _Indentation = require("./Indentation");

var _Indentation2 = _interopRequireDefault(_Indentation);

var _InlineBlock = require("./InlineBlock");

var _InlineBlock2 = _interopRequireDefault(_InlineBlock);

var _Params = require("./Params");

var _Params2 = _interopRequireDefault(_Params);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Formatter = function () {
    /**
     * @param {Object} cfg
     *   @param {Object} cfg.indent
     *   @param {Object} cfg.params
     * @param {Tokenizer} tokenizer
     */
    function Formatter(cfg, tokenizer) {
        _classCallCheck(this, Formatter);

        this.cfg = cfg || {};
        this.indentation = new _Indentation2["default"](this.cfg.indent);
        this.inlineBlock = new _InlineBlock2["default"]();
        this.params = new _Params2["default"](this.cfg.params);
        this.tokenizer = tokenizer;
        this.previousReservedWord = {};
        this.tokens = [];
        this.index = 0;
    }

    /**
     * Formats whitespaces in a SQL string to make it easier to read.
     *
     * @param {String} query The SQL query string
     * @return {String} formatted query
     */


    Formatter.prototype.format = function format(query) {
        this.tokens = this.tokenizer.tokenize(query);
        var formattedQuery = this.getFormattedQueryFromTokens();

        return formattedQuery.trim();
    };

    Formatter.prototype.getFormattedQueryFromTokens = function getFormattedQueryFromTokens() {
        var _this = this;

        var formattedQuery = "";

        this.tokens.forEach(function (token, index) {
            _this.index = index;
            if (token.type === _tokenTypes2["default"].WHITESPACE) {
                // ignore (we do our own whitespace formatting)
            } else if (token.type === _tokenTypes2["default"].LINE_COMMENT) {
                formattedQuery = _this.formatLineComment(token, formattedQuery);
                // formattedQuery = this.formatWithSpaces(token, formattedQuery);
            } else if (token.type === _tokenTypes2["default"].BLOCK_COMMENT) {
                formattedQuery = _this.formatBlockComment(token, formattedQuery);
            } else if (token.type === _tokenTypes2["default"].RESERVED_TOPLEVEL) {
                formattedQuery = _this.formatToplevelReservedWord(token, formattedQuery);
                _this.previousReservedWord = token;
            } else if (token.type === _tokenTypes2["default"].RESERVED_NEWLINE) {
                formattedQuery = _this.formatNewlineReservedWord(token, formattedQuery);
                if (/^ADD JAR/i.test(token.value) || /^SET/i.test(token.value)) {
                    _this.indentation.setWhiteSpace(false);
                }
                _this.previousReservedWord = token;
            } else if (token.type === _tokenTypes2["default"].RESERVED_TOPLEVEL_INLINE) {
                formattedQuery = _this.formatToplevelInLineReservedWord(token, formattedQuery);
                _this.previousReservedWord = token;
            } else if (token.type === _tokenTypes2["default"].UNION_WORDS) {
                formattedQuery = _this.formatUnionWords(token, formattedQuery);
                _this.previousReservedWord = token;
            } else if (token.type === _tokenTypes2["default"].RESERVED) {
                if (/^BETWEEN/i.test(token.value)) {
                    _this.indentation.setNoNewLine();
                }
                formattedQuery = _this.formaReserverdWords(token, formattedQuery);
                _this.previousReservedWord = token;
            } else if (token.type === _tokenTypes2["default"].OPEN_PAREN) {
                formattedQuery = _this.formatOpeningParentheses(token, formattedQuery);
            } else if (token.type === _tokenTypes2["default"].CLOSE_PAREN) {
                formattedQuery = _this.formatClosingParentheses(token, formattedQuery);
            } else if (token.type === _tokenTypes2["default"].PLACEHOLDER) {
                formattedQuery = _this.formatPlaceholder(token, formattedQuery);
            } else if (token.value === ",") {
                formattedQuery = _this.formatComma(token, formattedQuery);
            } else if (token.value === ":") {
                formattedQuery = _this.formatWithoutSpaces(token, formattedQuery);
            } else if (token.value === ".") {
                formattedQuery = _this.formatWithoutSpaces(token, formattedQuery);
            } else if (token.value === ";") {
                _this.indentation.setWhiteSpace(true);
                formattedQuery = _this.endCodeBlock(token, formattedQuery);
            } else {
                formattedQuery = _this.formatWithSpaces(token, formattedQuery);
            }
        });
        return formattedQuery;
    };

    Formatter.prototype.formatLineComment = function formatLineComment(token, query) {
        var follow = this.followNonWhitespaceTokenIndex();
        if (follow.token.value === ",") {
            var commaToken = this.tokens[follow.index];
            this.tokens[follow.index] = token;
            this.tokens[this.index] = commaToken;
            return this.formatComma(commaToken, query);
        }

        if (this.previousNonWhitespaceToken().type === _tokenTypes2["default"].UNION_WORDS) {

            query = (0, _trimEnd2["default"])(query) + " " + token.value;
            this.indentation.setNoTrimEnd();
            return query;
        } else if (this.previousNonWhitespaceToken().value !== ";" && this.previousNonWhitespaceToken().type !== _tokenTypes2["default"].LINE_COMMENT) {
            return (0, _trimEnd2["default"])(query) + " " + this.addNewline(token.value);
        } else {
            return this.addNewline(query) + this.addNewline(token.value);
        }
    };

    Formatter.prototype.formatBlockComment = function formatBlockComment(token, query) {
        return this.addNewline(this.addNewline(query) + this.indentComment(token.value));
    };

    Formatter.prototype.indentComment = function indentComment(comment) {
        return comment.replace(/\n/g, "\n" + this.indentation.getIndent());
    };

    Formatter.prototype.formatToplevelReservedWord = function formatToplevelReservedWord(token, query) {
        this.indentation.decreaseTopLevel();
        if (this.previousNonWhitespaceToken().type !== _tokenTypes2["default"].OPEN_PAREN) {
            query = this.addNewline(query);
        } else {
            query = (0, _trimEnd2["default"])(query) + " ";
        }
        this.indentation.increaseToplevel();

        query += this.equalizeWhitespace(token.value.toLowerCase());
        return this.addNewline(query);
    };

    Formatter.prototype.formatToplevelInLineReservedWord = function formatToplevelInLineReservedWord(token, query) {
        this.indentation.decreaseTopLevel();

        query = this.addNewline(query);

        this.indentation.increaseToplevel();

        query += this.equalizeWhitespace(token.value.toLowerCase());
        return query + " ";
    };

    Formatter.prototype.formatUnionWords = function formatUnionWords(token, query) {
        this.indentation.setNoTrimEnd();
        query = query + "\n" + this.indentation.getIndent();
        this.indentation.decreaseTopLevel();
        this.indentation.setNoTrimEnd();
        query = query + "\n" + this.indentation.getIndent();

        query = query + token.value.toLowerCase() + "\n";
        this.indentation.setNoTrimEnd();

        return query;
    };

    Formatter.prototype.formatNewlineReservedWord = function formatNewlineReservedWord(token, query) {
        if (this.indentation.getStartNewLine()) {
            return this.addNewline(query) + this.equalizeWhitespace(token.value.toLowerCase()) + " ";
        } else {
            return query + this.equalizeWhitespace(token.value.toLowerCase()) + " ";
        }
    };

    // Replace any sequence of whitespace characters with single space


    Formatter.prototype.equalizeWhitespace = function equalizeWhitespace(string) {
        return string.replace(/\s+/g, " ");
    };

    // Opening parentheses increase the block indent level and start a new line


    Formatter.prototype.formatOpeningParentheses = function formatOpeningParentheses(token, query) {
        // Take out the preceding space unless there was whitespace there in the original query
        // or another opening parens or line comment
        if (this.previousNonWhitespaceToken().type === _tokenTypes2["default"].RESERVED_TOPLEVEL_INLINE || this.previousNonWhitespaceToken().value === "," && token.value.toLowerCase() === "case") {
            query = this.addNewline(query) + token.value;
            this.indentation.increaseBlockLevel();
            return query;
        } else {
            var preserveWhitespaceFor = [_tokenTypes2["default"].WHITESPACE, _tokenTypes2["default"].OPEN_PAREN, _tokenTypes2["default"].LINE_COMMENT];
            if (!preserveWhitespaceFor.includes(this.previousToken().type)) {
                query = (0, _trimEnd2["default"])(query);
            }
            query += token.value;

            this.inlineBlock.beginIfPossible(this.tokens, this.index);

            if (!this.inlineBlock.isActive()) {
                this.indentation.increaseBlockLevel();
                query = this.addNewline(query);
            }
            return query;
        }
    };

    // Closing parentheses decrease the block indent level


    Formatter.prototype.formatClosingParentheses = function formatClosingParentheses(token, query) {
        if (this.inlineBlock.isActive()) {
            this.inlineBlock.end();
            return this.formatWithSpaceAfter(token, query);
        } else {
            this.indentation.decreaseBlockLevel();
            return this.formatWithSpaces(token, this.addNewline(query));
        }
    };

    Formatter.prototype.formatPlaceholder = function formatPlaceholder(token, query) {
        return query + this.params.get(token) + " ";
    };

    Formatter.prototype.endCodeBlock = function endCodeBlock(token, query) {
        query = query + token.value + "\n";
        this.indentation.setNoTrimEnd();
        this.indentation.decreaseBlockLevel();
        return query;
    };

    // Commas start a new line (unless within inline parentheses or SQL "LIMIT" clause)


    Formatter.prototype.formatComma = function formatComma(token, query) {
        query = this.trimTrailingWhitespace(query) + token.value + " ";

        if (this.inlineBlock.isActive()) {
            return query;
        } else if (/^LIMIT$/i.test(this.previousReservedWord.value)) {
            return query;
        } else {
            return this.addNewline(query);
        }
    };

    Formatter.prototype.formatWithSpaceAfter = function formatWithSpaceAfter(token, query) {
        return this.trimTrailingWhitespace(query) + token.value + " ";
    };

    Formatter.prototype.formatWithoutSpaces = function formatWithoutSpaces(token, query) {
        return this.trimTrailingWhitespace(query) + token.value;
    };

    Formatter.prototype.formaReserverdWords = function formaReserverdWords(token, query) {
        if (this.indentation.getWhiteSpace()) {
            return query + token.value.toLowerCase() + " ";
        } else {
            return query + token.value.toLowerCase();
        }
    };

    Formatter.prototype.formatWithSpaces = function formatWithSpaces(token, query) {
        if (this.indentation.getWhiteSpace()) {
            return query + token.value + " ";
        } else {
            return query + token.value;
        }
    };

    Formatter.prototype.addNewline = function addNewline(query) {
        if (this.indentation.getTrimEnd()) {
            return (0, _trimEnd2["default"])(query) + "\n" + this.indentation.getIndent();
        } else {
            return query + "\n" + this.indentation.getIndent();
        }
    };

    Formatter.prototype.trimTrailingWhitespace = function trimTrailingWhitespace(query) {
        if (this.previousNonWhitespaceToken().type === _tokenTypes2["default"].LINE_COMMENT) {
            return (0, _trimEnd2["default"])(query) + "\n";
        } else {
            return (0, _trimEnd2["default"])(query);
        }
    };

    Formatter.prototype.previousNonWhitespaceToken = function previousNonWhitespaceToken() {
        var n = 1;
        while (this.previousToken(n).type === _tokenTypes2["default"].WHITESPACE) {
            n++;
        }
        return this.previousToken(n);
    };

    Formatter.prototype.previousToken = function previousToken() {
        var offset = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 1;

        return this.tokens[this.index - offset] || {};
    };

    Formatter.prototype.followNonWhitespaceTokenIndex = function followNonWhitespaceTokenIndex() {
        var n = 1;
        while (this.followToken(n).type === _tokenTypes2["default"].WHITESPACE) {
            n++;
        }
        return { token: this.followToken(n), index: this.index + n };
    };

    Formatter.prototype.followToken = function followToken() {
        var offset = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 1;

        return this.tokens[this.index + offset] || {};
    };

    return Formatter;
}();

exports["default"] = Formatter;
module.exports = exports["default"];